import jmespath


class URDBPath:

    def __init__(self, expression, **meta):
        """Assign an expression to the URDB json path.

        Args:
            expression (str): A valid jmespath expression.
            **meta: key-value pairs associated with the expression.
        """
        self.expression = jmespath.compile(expression)
        self.metadata = meta

    def search(self, data):
        """Proxy method for compiled jmespath search.
        """
        return self.expression.search(data)


class URDBMeta:

    name = URDBPath("name")

    description = URDBPath("description")


class Energy:

    path = URDBPath("energyratestructure[].rate")

    weekend_schedule = URDBPath("energyweekendschedule", is_weekday=False)

    weekday_schedule = URDBPath("energyweekdayschedule", is_weekday=True)

    attrs = URDBPath("energyattrs")


class Demand:

    path = URDBPath("demandratestructure[].rate")

    weekend_schedule = URDBPath("demandweekendschedule", is_weekday=False)

    weekday_schedule = URDBPath("demandweekdayschedule", is_weekday=True)

    attrs = URDBPath("demandattrs")

class FlatDemand:

    path = URDBPath("flatdemandstructure[].rate")

    schedule = URDBPath("flatdemandmonths")


class Meter:
    
    path = URDBPath("fixedchargefirstmeter")

    charge_unit = URDBPath("fixedchargeunits")


import requests
import pandas as pd

from dataclasses import dataclass, field
from urllib.parse import urljoin

from ..abstract import AbstractRate
from .schemas import URDBMeta, Energy, Demand, FlatDemand, Meter


root = "https://api.openei.org"


class UnknownRateStructure(Exception):
    pass


@dataclass
class RateSchedule(AbstractRate):

    apikey: str = field(repr=False)
    openei_schedule_id: str

    _rate: "typing.Any" = field(init=False, repr=False, default=None)
    _energy: "typing.Any" = field(init=False, repr=False, default=None)
    _demand: "typing.Any" = field(init=False, repr=False, default=None)
    _flatdemand: "typing.Any" = field(init=False, repr=False, default=None)
    _meter: "typing.Any" = field(init=False, repr=False, default=None)

    url = urljoin(root, "utility_rates")

    def __post_init__(self):
        self.name = URDBMeta.name.search(self.rate)
        self.description = URDBMeta.description.search(self.rate)

    @property
    def rate(self):
        """Get OpenEI Rate.
        """
        if self._rate is None:
            self._rate = self.load_rate()
        return self._rate

    def load_rate(self):
        params = {
            "api_key": self.apikey,
            "getpage": self.openei_schedule_id,
            "format": "json",
            # Unclear whether 'version' changes response schema.  7 at time of development.
            "version": "latest",
            "detail": "full",
        }
        rate = requests.get(self.url, params=params).json()

        # Python 3.8
        # assert (n := len(rate["items"])) == 1, f"Expected 1 rate, found {n}"
        n = len(rate["items"])
        assert n == 1, f"Expected 1 rate, found {n}"
        return rate["items"][0]

    @property
    def energy(self):
        """Create a month-hour schedule for $/kWh charges.

        Returns:
            DataFrame of rate schedules with granularity of Month, hour of day and weekday.
            Length should be 12 * 24 * 2

        Schema:

            month            int64
            hour            object
            schedule_id      int64
            rate           float64
            is_weekday        bool
        """
        if self._energy is None:
            self._energy = self._parse_tou_schedule(Energy)
        return self._energy

    @property
    def demand(self):
        if self._demand is None:
            self._demand = self._parse_tou_schedule(Demand)
        return self._demand

    def _parse_tou_schedule(self, schema):
        """
        """
        rate = pd.Series(schema.path.search(self.rate))
        # Create empty dataframe to hold rate schedules.
        rate_schedule = pd.DataFrame()
        for _s in (schema.weekend_schedule, schema.weekday_schedule,):

            # Each row represents 1 month
            schedule = pd.DataFrame(_s.search(self.rate)).rename_axis(
                "month").reset_index()
            schedule = pd.melt(schedule, id_vars=[
                "month"], var_name="hour", value_name="schedule_id")

            # Increment 0-indexed months to 1-indexed months
            schedule["month"] += 1

            # Map Rates
            schedule["rate"] = schedule["schedule_id"].map(rate)

            # Add to master rate schedule
            rate_schedule = rate_schedule.append(
                schedule.assign(**_s.metadata), ignore_index=True)

        return rate_schedule

    @property
    def flatdemand(self):
        """Flat demand charges, applicable over a monthly billing period.

        Flat demand charges have only a single, flat value per month.

        Schema:

            month            int64
            schedule_id      int64
            rate           float64
        """
        if self._flatdemand is None:
            self._flatdemand = self._parse_flatdemand_rates()
        return self._flatdemand

    def _parse_flatdemand_rates(self):
        """Get flat demand charges.

        Flat Demand charges are monthly $/kW charges for the maximum 15-min averages.
        """
        rate = pd.Series(FlatDemand.path.search(self.rate))

        schedule = pd.Series(
            FlatDemand.schedule.search(
                self.rate)
        ).rename_axis("month").rename("schedule_id").reset_index()
        schedule["month"] += 1
        schedule["rate"] = schedule["schedule_id"].map(rate)
        return schedule

    def _parse_charge_period(self, attrs, period_attr):
        """

        Args:
            attrs (list[dict]): List of dictionaries from openEI.
            period_attr (str): One of "season" or "period"
        """
        component_parser = {
            "season": lambda x: x.split(":", 1)[0].split("-")[1].title(),
            "period": lambda x: x.split(":", 1)[1].title(),
        }
        try:
            parser = component_parser[period_attr]
        except KeyError:
            raise Exception("Invalid period attribute")

        components = {}

        for rate_attribute in attrs:
            if not len(rate_attribute) == 1:
                raise UnknownRateStructure("Expected 1 attribute per rate attribute list.")
            slug, schedule_id = list(rate_attribute.items())[0]
            # Filter for only seasons/periods.
            # e.g. "TOU-winter:Off-Peak"
            if ":" not in slug:
                continue
            # 0-index the schedule id to be internally consistent
            schedule_id = int(schedule_id) - 1
            period = parser(slug)
            components[schedule_id] = period
        return components

    @property
    def seasons(self):
        energy_attrs = Energy.attrs.search(self.rate)
        if energy_attrs:
            return self._parse_charge_period(energy_attrs, "season") 
        else:
            return {i: f"Season-{i}" for i, rate in Energy(Demand.path.search(self.rate))}

    @property
    def energy_periods(self):
        energy_attrs = Energy.attrs.search(self.rate)
        if energy_attrs:
            return self._parse_charge_period(energy_attrs, "period")
        else:
            return {i: f"Period-{i}" for i, rate in Energy(Demand.path.search(self.rate))}

    @property
    def demand_periods(self):
        """Mapping between schedule ids in rates and human readable period names.
        """
        demand_attrs = Demand.attrs.search(self.rate)
        # Not every rate has descriptive attributes for the demand schedule
        if demand_attrs:
            demand_periods = self._parse_charge_period(demand_attrs, "period")
        else:
            demand_periods = {i: f"Period-{i}" for i, rate in enumerate(Demand.path.search(self.rate))}
        return demand_periods

    @property
    def meter_charge_unit(self):
        rate = self.rate
        return Meter.charge_unit.search(rate)

    @property
    def meter(self):
        rate = self.rate
        return Meter.path.search(rate)


from abc import ABC, abstractmethod


class AbstractRate(ABC):

    @property
    @abstractmethod
    def energy(self):
        pass

    @property
    @abstractmethod
    def demand(self):
        pass

    @property
    @abstractmethod
    def flatdemand(self):
        pass



import pandas as pd
from dataclasses import dataclass

from .calculator import SingleSite


@dataclass
class MonthlyBillReport:
    """Create a monthly bill report with support for TOU-specific rates.
    """

    load: pd.Series
    name: str
    calculator: SingleSite

    def __post_init__(self):
        self.load = self.load.copy()
        self.load.name = self.load.name or "kW"
        self.load.index.name = self.load.index.name or "timestamp"

    @property
    def seasonal_load(self):
        rate_grain = [
            "is_weekday",
            "month",
            "hour",
        ]
        load = self.load.reset_index()

        # Load is a function of month, hour, is_weekday
        load_grain = [
            ~load[self.load.index.name].dt.dayofweek.isin([5, 6]),
            load[self.load.index.name].dt.month,
            load[self.load.index.name].dt.hour,
        ]

        load = load.merge(
            self.calculator.schedule.energy,
            right_on=rate_grain,
            left_on=load_grain,
            how="left",
        )
        load["season"] = load["schedule_id"].map(self.calculator.schedule.seasons)

        return load[[self.load.name, self.load.index.name, "season"]].set_index(
            self.load.index.name
        )

    @property
    def demand(self):
        """
        Aggregates all demand charges and values to a single value.
        """
        cols = {"cost": "Demand ($)", self.load.name: "Demand (kW)"}
        flatdemand_charges = self.calculator.calculate_flatdemand_charges(
            self.load
        ).rename(columns=cols)
        flat_charges = flatdemand_charges[list(cols.values())]

        cols = {"cost": "Demand ($)", self.load.name: "Demand (kW)"}
        peak_charges = (
            self.calculator.calculate_demand_charges(self.load)
            .rename(columns=cols)
            .rename(columns=self.calculator.schedule.demand_periods, level=1)
        )

        for dup_col in peak_charges.columns[peak_charges.columns.duplicated()]:
            S = peak_charges[dup_col].sum(axis=1)
            peak_charges.drop(columns=dup_col, inplace=True)
            peak_charges[dup_col] = S

        peak_charges = peak_charges[list(cols.values())]
        peak_charges.columns = peak_charges.columns.map(" - ".join)
        return peak_charges.join(flat_charges)

    @property
    def energy(self):
        energy_cols = {"cost": "Energy ($)", "kWh": "Energy (kWh)"}
        energy_charges = self.calculator.calculate_energy_charges(
            self.load.rename("kWh")
        ).rename(columns=energy_cols)
        return energy_charges[list(energy_cols.values())].resample("M").sum()

    @property
    def meter(self):
        meter_charges = self.calculator.calculate_meter_charges(self.load).rename(
            "Meter ($)"
        )
        return meter_charges

    @property
    def monthly(self):
        monthly = self.energy.join(self.demand).join(self.meter)
        billing_cols = monthly.columns[monthly.columns.str.contains(r"\(\$\)")]
        monthly["Total ($)"] = monthly[billing_cols].sum(axis=1).rename("Total ($)")
        return monthly.rename_axis("Month")

    @property
    def annual(self):
        return self.monthly.sum(axis=0).rename(self.name)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ubc', 'ubc.rates', 'ubc.rates.openei']

package_data = \
{'': ['*']}

install_requires = \
['altair>=4.1.0,<5.0.0',
 'dynaconf>=2.2.3,<3.0.0',
 'jmespath>=0.2.1,<0.3.0',
 'pandas>=1.0.3,<2.0.0',
 'requests>=2.23.0,<3.0.0']

setup_kwargs = {
    'name': 'ubc',
    'version': '0.0.12',
    'description': 'Utility Bill Calculator',
    'long_description': None,
    'author': 'Thomas Tu',
    'author_email': 'thomasthetu@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from pandas import Timedelta, NA

from ubc.rates.abstract import AbstractRate


DELTA_HOUR = Timedelta("1 hour")


class UnknownRateStructure(Exception):
    pass


@dataclass
class SingleSite:
    """Single-Site Bill Calculator
    """

    schedule: AbstractRate

    def calculate_energy_charges(self, load):
        """
        Args:
            load (pd.Series):  Time indexed series of load data in kWh.
        """
        load = load.copy()
        load.rename(load.name or "kWh", inplace=True)
        load.rename_axis(load.index.name or "timestamp", inplace=True)

        load_grain = [
            ~load.index.dayofweek.isin([5, 6]),
            load.index.month,
            load.index.hour,
        ]
        schedule_grain = ["is_weekday", "month", "hour"]

        # Assign rates to the provided load data.
        rates = self.schedule.energy.merge(
            load.reset_index(), right_on=load_grain, left_on=schedule_grain, how="right"
        )[[load.index.name, load.name, "rate"]]

        rates["cost"] = rates[load.name] * rates["rate"]
        rates.rename(columns={"rate": f"{load.name}_rate"}, inplace=True)
        return rates.set_index(load.index.name).sort_index()

    def calculate_demand_charges(self, load):
        """
        Args:
            load (pd.Series):  Time indexed series of load data in kW.
        """
        load = load.copy()
        load.rename(load.name or "kW", inplace=True)
        load.rename_axis(load.index.name or "timestamp", inplace=True)

        # Resample to 15min intervals and then upsample to month level taking maximum demand
        load = load.resample("15 min").mean().fillna(method="pad").resample("H").max()

        load_grain = [
            ~load.index.dayofweek.isin([5, 6]),
            load.index.month,
            load.index.hour,
        ]
        schedule_grain = ["is_weekday", "month", "hour"]

        # Assign rates to the provided load data.
        demand_rates = self.schedule.demand

        rates = demand_rates.merge(
            load.reset_index(), right_on=load_grain, left_on=schedule_grain, how="right"
        )[[load.index.name, load.name, "rate", "schedule_id"]]
        # Calculate max demand over 15 min. period for each month

        if demand_rates.empty:
            rates["schedule_id"] = rates["schedule_id"].fillna(int(0))
            rates["rate"] = NA

        rates[f"cost"] = rates[load.name] * rates["rate"]
        rates = rates.set_index(load.index.name).sort_index()
        rates.rename(columns={"rate": f"{load.name}_rate"}, inplace=True)
        # Finally, only take the maximum demand charge in each billing period for each schedule ID
        # Each schedule ID represents a different demand charge type (part peak, full peak)
        rates = rates.pivot(
            columns="schedule_id", values=[load.name, f"{load.name}_rate", f"cost"]
        )
        return rates.resample("M").max()

    def calculate_flatdemand_charges(self, load):
        """
        Args:
            load (pd.Series):  Time indexed series of load data in kW.
        """
        load = load.copy()
        load.rename(load.name or "kW", inplace=True)
        load.rename_axis(load.index.name or "timestamp", inplace=True)

        # Resample to 15min intervals and then upsample to month level taking maximum demand
        load = load.resample("15 min").mean().fillna(method="pad").resample("M").max()

        load_grain = [load.index.month]
        schedule_grain = ["month"]

        # Assign rates to the provided load data.
        rates = self.schedule.flatdemand.merge(
            load.reset_index(), right_on=load_grain, left_on=schedule_grain, how="right"
        )[[load.index.name, load.name, "rate"]]

        # Calculate max demand over 15 min. period for each month
        rates["cost"] = rates[load.name] * rates["rate"]
        rates.rename(columns={"rate": f"{load.name}_rate"}, inplace=True)
        return rates.set_index(load.index.name).sort_index()

    def calculate_meter_charges(self, load, charge_type=None):
        """
        Args:
            load (pd.Series): timestamp/series of demand data.
            charge_type (str): if the 
        """
        charge_unit = self.schedule.meter_charge_unit
        if charge_unit == "$/day":
            # Meter charges exist per-diem, the actual aggregate load value doesn't matter.
            # Aggregate to dailies, then count up the months.
            N = load.resample("D").nearest().resample("M").count()
        elif charge_unit == "$/month":
            # if meter charges exist per-month, we simply have one charge per month in our index
            # So normalize to 1!
            N = load.resample("D").nearest().resample("M").count()
            N = N / N
        else:
            raise UnknownRateStructure("Unknown meter charge unit: {}".format())
        cost = N * self.schedule.meter
        return cost.rename("cost")

    def calculate_total(self, load_kw):
        """
        Args:
            load_kw (pd.Series): timestamp series of demand data.
        """

        interval = load_kw.index.freq.delta / DELTA_HOUR
        energy_charges = (
            self.calculate_energy_charges(load_kw * interval)["cost"]
            .resample("M")
            .sum()
        )
        demand_charges = self.calculate_demand_charges(load_kw)["cost"].sum(axis=1)
        flatdemand_charges = self.calculate_flatdemand_charges(load_kw)["cost"]
        meter_charges = self.calculate_meter_charges(load_kw)
        return (
            meter_charges + energy_charges + demand_charges + flatdemand_charges
        ).rename("total_cost")
